import { Injectable } from '@angular/core';
import {
  Client,
  ClientNotificationCodec,
  WebSocketClientRWC,
  WebSocketClientRWCConfig,
  JSONClientCodec
} from '@overflow/rpc-js';
import { Subscription } from 'rxjs';
import { TypeUtil, Class, Annotation, PropertyKeyType, Method } from '@overflow/core-js';
import { RPCSubscriberDecoratorAttribute } from '@overflow/commons/ui/decorator/RPCSubscriber';

export const requesterID = 'scannerUser';

export interface SubscriberMethod {
  className: PropertyKeyType;
  methodName: PropertyKeyType;
  parameterTypes: string[];

  method: Method;
  instance: any;
}

export class SubscriberParameterError extends Error {
  public constructor(message?: string) {
    super(message);
    Object.setPrototypeOf(this, new.target.prototype);
  }
}

export class SubscriberExistError extends Error {
  public constructor(message?: string) {
    super(message);
    Object.setPrototypeOf(this, new.target.prototype);
  }
}


const config: WebSocketClientRWCConfig = {
  url: 'ws://localhost:60000/scanner?requesterID=' + requesterID
};
const rwc = new WebSocketClientRWC(config);
const codec = new JSONClientCodec();

@Injectable({
  providedIn: 'root'
})
export class ProbeService extends Client {
  private notiSubscription: Subscription;

  private subscriberMap: Map<any, string[]>;
  private subscriberMethodMap: Map<string, SubscriberMethod[]>;

  public constructor() {
    super(codec, rwc);

    this.subscriberMap = new Map();
    this.subscriberMethodMap = new Map();

    this.connect();

    this.notiSubscription = this.notification()
      .subscribe((notiCodec: ClientNotificationCodec) => {
        const method = notiCodec.method();
        const params = notiCodec.params();

        const subscriberMethods: SubscriberMethod[] = this.subscriberMethodMap.get(method);
        if (undefined === subscriberMethods) {
          console.warn(`Subscriber for method[${method}] is not exist`);
          return;
        }
        subscriberMethods.forEach((subscriberMethod) => {
          try {
            const args = this.converParams(params, subscriberMethod.parameterTypes);
            subscriberMethod.method.invoke(subscriberMethod.instance, ...args);
          } catch (error) {
            console.error(error);
          }
        });
      });
    this.connectionStatus().subscribe(isConnected => {
      if (!isConnected) {
        console.log('disconnected');
      } else {
        console.log('connected');
      }
    });

  }

  /**
   * subscribeNotification
   */
  public subscribeNotification(subscriber: any) {
    if (this.subscriberMap.has(subscriber)) {
      throw new SubscriberExistError();
    }

    const type = TypeUtil.getType(subscriber);

    const clazz = Class.forType(type);
    if (undefined === clazz) {
      console.log(`Type[${subscriber.name}] is not decorated type`);
      return;
    }

    const subscriberMethodNames: string[] = [];
    this.subscriberMap.set(subscriber, subscriberMethodNames);

    const methods = clazz.getMethods();
    methods.forEach((method, propertyKey) => {
      const annon = method.getAnnotation<Annotation<RPCSubscriberDecoratorAttribute>>(Annotation);
      if (undefined === annon) {
        return;
      }

      const subscriberMethodName = annon.attribute.method;
      subscriberMethodNames.push(subscriberMethodName);
      let subscriberMethods: SubscriberMethod[] = this.subscriberMethodMap.get(subscriberMethodName);
      if (undefined === subscriberMethods) {
        subscriberMethods = [];
        this.subscriberMethodMap.set(subscriberMethodName, subscriberMethods);
      }

      const paramTypes = this.getParamTypes(method);

      const subscriberMethod: SubscriberMethod = {
        className: clazz.getName(),
        methodName: method.getName(),
        parameterTypes: paramTypes,
        method: method,
        instance: subscriber,
      };

      subscriberMethods.push(subscriberMethod);
    });
  }

  /**
   * subscribeNotification
   */
  public unsubscribeNotification(subscriber: any) {
    if (!this.subscriberMap.has(subscriber)) {
      return;
    }

    const subscriberMethodNames = this.subscriberMap.get(subscriber);
    this.subscriberMap.delete(subscriber);
    subscriberMethodNames.forEach(methodName => {
      if (!this.subscriberMethodMap.has(methodName)) {
        return;
      }
      const subscriberMethods = this.subscriberMethodMap.get(methodName);
      subscriberMethods.some((subscriberMethod, index): boolean => {
        if (subscriberMethod.instance === subscriber) {
          delete subscriberMethods[index];
        }
        return false;
      });
    });
  }

  private getParamTypes(method: Method): string[] {
    if (undefined === method || null === method || 0 === method.getParameterCount()) {
      return [];
    }

    const parameters = method.getParameters();
    const results: string[] = [];
    for (let indexI = 0; indexI < parameters.length; indexI++) {
      const paramType = parameters[indexI].getType();
      results.push(paramType.name);
    }
    return results;
  }

  private converParams(params: string[], paramTypes: string[]): any[] {
    const results: any[] = [];

    if (undefined === params || null === params || 0 === params.length) {
      return results;
    }
    if (undefined === paramTypes || null === paramTypes || 0 === paramTypes.length) {
      return results;
    }
    if (params.length !== paramTypes.length) {
      throw new SubscriberParameterError(`Count is not same from server[${params.length}] and method[${paramTypes.length}]`);
    }
    for (let indexI = 0; indexI < params.length; indexI++) {
      const param = params[indexI];
      const type = paramTypes[indexI];
      switch (type) {
        case 'Object':
        case 'Array':
        case 'Map':
          results.push(JSON.parse(param));
          break;
        case 'String':
          results.push(param);
          break;
        case 'Number':
          results.push(Number(param));
          break;
        case 'Boolean':
          results.push(Boolean(param));
          break;
        case 'Date':
          results.push(new Date(Number(param)));
          break;
        case 'Function':
          throw new SubscriberParameterError(`Function type [${indexI}] is not allowed`);
        default:
          throw new SubscriberParameterError(`${type} type parameter[${indexI}] is not allowed`);
      }
    }
    return results;
  }


}
